@@ -9,15 +9,13 @@ import asyncio
 import warnings
 import concurrent.futures
 
-from ._handles import Handle, TimerHandle
-from ._util import run_aio_future, run_aio_generator
-from ._deprecate import deprecated, deprecated_alias
-from . import _util
+from ._handles import ScopedHandle, AsyncHandle
+from ._util import run_aio_future
 
 from selectors import _BaseSelectorImpl, EVENT_READ, EVENT_WRITE
 
 try:
-    from trio.hazmat import wait_for_child
+    from trio.lowlevel import wait_for_child
 except ImportError:
     from ._child import wait_for_child
 
@@ -27,11 +25,11 @@ logger = logging.getLogger(__name__)
 _mswindows = (sys.platform == "win32")
 
 try:
-    _wait_readable = trio.hazmat.wait_readable
-    _wait_writable = trio.hazmat.wait_writable
+    _wait_readable = trio.lowlevel.wait_readable
+    _wait_writable = trio.lowlevel.wait_writable
 except AttributeError:
-    _wait_readable = trio.hazmat.wait_socket_readable
-    _wait_writable = trio.hazmat.wait_socket_writable
+    _wait_readable = trio.lowlevel.wait_socket_readable
+    _wait_writable = trio.lowlevel.wait_socket_writable
 
 
 class _Clear:
@@ -39,28 +37,6 @@ class _Clear:
         pass
 
 
-def _h_raise(handle, exc):
-    """
-    Convince a handle to raise an error.
-
-    trio-asyncio enhanced handles have a method to do this
-    but asyncio's native handles don't. Thus we need to fudge things.
-    """
-    if hasattr(handle, '_raise'):
-        handle._raise(exc)
-        return
-
-    def _raise(exc):
-        raise exc
-
-    cb, handle._callback = handle._callback, _raise
-    ar, handle._args = handle._args, (exc,)
-    try:
-        handle._run()
-    finally:
-        handle._callback, handle._args = cb, ar
-
-
 class _TrioSelector(_BaseSelectorImpl):
     """A selector that hooks into a ``TrioEventLoop``.
 
@@ -241,25 +217,6 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
         finally:
             sniffio.current_async_library_cvar.reset(t)
 
-    async def __run_trio(self, h):
-        """Helper for copying the result of a Trio task to an asyncio future"""
-        f, proc, *args = h._args
-        if f.cancelled():  # pragma: no cover
-            return
-        try:
-            with trio.CancelScope() as scope:
-                h._scope = scope
-                res = await proc(*args)
-            if scope.cancelled_caught:
-                f.cancel()
-                return
-        except BaseException as exc:
-            if not f.cancelled():  # pragma: no branch
-                f.set_exception(exc)
-        else:
-            if not f.cancelled():  # pragma: no branch
-                f.set_result(res)
-
     def trio_as_future(self, proc, *args):
         """Start a new Trio task to run ``await proc(*args)`` asynchronously.
         Return an `asyncio.Future` that will resolve to the value or exception
@@ -292,14 +249,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
           an `asyncio.Future` which will resolve to the result of the call to *proc*
         """
         f = asyncio.Future(loop=self)
-        h = Handle(
-            self.__run_trio, (
-                f,
-                proc,
-            ) + args, self, context=None, is_sync=None
-        )
-        self._queue_handle(h)
-        f.add_done_callback(h._cb_future_cancel)
+        self._queue_handle(AsyncHandle(proc, args, self, result_future=f))
         return f
 
     def run_trio_task(self, proc, *args):
@@ -314,7 +264,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
         Returns:
           an `asyncio.Handle` which can be used to cancel the background task
         """
-        return self._queue_handle(Handle(proc, args, self, is_sync=False))
+        return self._queue_handle(AsyncHandle(proc, args, self))
 
     # Callback handling #
 
@@ -331,7 +281,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
     def _call_soon(self, *arks, **kwargs):
         raise RuntimeError("_call_soon() should not have been called")
 
-    def call_later(self, delay, callback, *args, context=None):
+    def call_later(self, delay, callback, *args, **context):
         """asyncio's timer-based delay
 
         Note that the callback is a sync function.
@@ -342,36 +292,34 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
         """
         self._check_callback(callback, 'call_later')
         assert delay >= 0, delay
-        h = TimerHandle(delay + self.time(), callback, args, self, context=context, is_sync=True)
+        h = asyncio.TimerHandle(delay + self.time(), callback, args, self, **context)
         self._queue_handle(h)
         return h
 
-    def call_at(self, when, callback, *args, context=None):
+    def call_at(self, when, callback, *args, **context):
         """asyncio's time-based delay
 
         Note that the callback is a sync function.
         """
         self._check_callback(callback, 'call_at')
-        return self._queue_handle(
-            TimerHandle(when, callback, args, self, context=context, is_sync=True)
-        )
+        return self._queue_handle(asyncio.TimerHandle(when, callback, args, self, **context))
 
-    def call_soon(self, callback, *args, context=None):
+    def call_soon(self, callback, *args, **context):
         """asyncio's defer-to-mainloop callback executor.
 
         Note that the callback is a sync function.
         """
         self._check_callback(callback, 'call_soon')
-        return self._queue_handle(Handle(callback, args, self, context=context, is_sync=True))
+        return self._queue_handle(asyncio.Handle(callback, args, self, **context))
 
-    def call_soon_threadsafe(self, callback, *args, context=None):
+    def call_soon_threadsafe(self, callback, *args, **context):
         """asyncio's thread-safe defer-to-mainloop
 
         Note that the callback is a sync function.
         """
         self._check_callback(callback, 'call_soon_threadsafe')
         self._check_closed()
-        h = Handle(callback, args, self, context=context, is_sync=True)
+        h = asyncio.Handle(callback, args, self, **context)
         self._token.run_sync_soon(self._q_send.send_nowait, h)
 
     # drop all timers
@@ -465,13 +413,13 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
 
         This is a Trio-flavored async function.
 
-        From asyncio, call ``await loop.run_trio(loop.synchronize)``
+        From asyncio, call ``await trio_as_aio(loop.synchronize)()``
         instead of ``await asyncio.sleep(0)`` if you need to process all
         queued callbacks.
 
         """
         w = trio.Event()
-        self._queue_handle(Handle(w.set, (), self, is_sync=True))
+        self._queue_handle(asyncio.Handle(w.set, (), self))
         await w.wait()
 
     # Signal handling #
@@ -488,7 +436,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
         self._check_signal(sig)
         if sig == signal.SIGKILL:
             raise RuntimeError("SIGKILL cannot be caught")
-        h = Handle(callback, args, self, context=None, is_sync=True)
+        h = asyncio.Handle(callback, args, self)
         assert sig not in self._signal_handlers, \
             "Signal %d is already being caught" % (sig,)
         self._orig_signals[sig] = signal.signal(sig, self._handle_sig)
@@ -516,7 +464,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
         ready for reading.
 
         This creates a new Trio task. You may want to use "await
-        :obj:`trio.hazmat.wait_readable` (fd)" instead, or
+        :obj:`trio.lowlevel.wait_readable` (fd)" instead, or
 
         :param fd: Either an integer (Unix file descriptor) or an object
                    with a ``fileno`` method providing one.
@@ -528,7 +476,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
 
     def _add_reader(self, fd, callback, *args):
         self._check_closed()
-        handle = Handle(callback, args, self, context=None, is_sync=True)
+        handle = ScopedHandle(callback, args, self)
         reader = self._set_read_handle(fd, handle)
         if reader is not None:
             reader.cancel()
@@ -547,20 +495,17 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
             self._selector.modify(fd, mask | EVENT_READ, (handle, writer))
             return reader
 
-    async def _reader_loop(self, fd, handle, task_status=trio.TASK_STATUS_IGNORED):
-        task_status.started()
-        with trio.CancelScope() as scope:
-            handle._scope = scope
+    async def _reader_loop(self, fd, handle):
+        with handle._scope:
             try:
-                while not handle._cancelled:  # pragma: no branch
+                while True:
                     await _wait_readable(fd)
-                    handle._call_sync()
+                    if handle._cancelled:
+                        break
+                    handle._run()
                     await self.synchronize()
             except Exception as exc:
-                _h_raise(handle, exc)
-                return
-            finally:
-                handle._scope = None
+                handle._raise(exc)
 
     # writing to a file descriptor
 
@@ -569,7 +514,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
         ready for writing.
 
         This creates a new Trio task. You may want to use "await
-        :obj:`trio.hazmat.wait_writable` (fd) instead, or
+        :obj:`trio.lowlevel.wait_writable` (fd) instead, or
 
         :param fd: Either an integer (Unix file descriptor) or an object
                    with a ``fileno`` method providing one.
@@ -583,7 +528,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
 
     def _add_writer(self, fd, callback, *args):
         self._check_closed()
-        handle = Handle(callback, args, self, context=None, is_sync=True)
+        handle = ScopedHandle(callback, args, self)
         writer = self._set_write_handle(fd, handle)
         if writer is not None:
             writer.cancel()
@@ -601,20 +546,17 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
             self._selector.modify(fd, mask | EVENT_WRITE, (reader, handle))
             return writer
 
-    async def _writer_loop(self, fd, handle, task_status=trio.TASK_STATUS_IGNORED):
-        with trio.CancelScope() as scope:
-            handle._scope = scope
-            task_status.started()
+    async def _writer_loop(self, fd, handle):
+        with handle._scope:
             try:
-                while not handle._cancelled:  # pragma: no branch
+                while True:
                     await _wait_writable(fd)
-                    handle._call_sync()
+                    if handle._cancelled:
+                        break
+                    handle._run()
                     await self.synchronize()
             except Exception as exc:
-                _h_raise(handle, exc)
-                return
-            finally:
-                handle._scope = None
+                handle._raise(exc)
 
     def autoclose(self, fd):
         """
@@ -670,8 +612,8 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
         if self._nursery is not None or not self._stopped.is_set():
             raise RuntimeError("You can't enter a loop twice")
         self._nursery = nursery
-        self._task = trio.hazmat.current_task()
-        self._token = trio.hazmat.current_trio_token()
+        self._task = trio.lowlevel.current_task()
+        self._token = trio.lowlevel.current_trio_token()
 
     async def _main_loop(self, task_status=trio.TASK_STATUS_IGNORED):
         """Run the loop by processing its event queue.
@@ -687,8 +629,19 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
         sniffio.current_async_library_cvar.set("asyncio")
 
         try:
-            while not self._stopped.is_set():
-                await self._main_loop_one()
+            # The shield here ensures that if the context surrounding
+            # the loop is cancelled, we keep processing callbacks
+            # until we reach the callback inserted by stop().
+            # That's important to maintain the asyncio invariant
+            # that everything you schedule before stop() will run
+            # before the loop stops. In order to be safe against
+            # deadlocks, it's important that the surrounding
+            # context ensure that stop() gets called upon a
+            # cancellation. (open_loop() does this indirectly
+            # by calling _main_loop_exit().)
+            with trio.CancelScope(shield=True):
+                while not self._stopped.is_set():
+                    await self._main_loop_one()
         except StopAsyncIteration:
             # raised by .stop_me() to interrupt the loop
             pass
@@ -717,7 +670,7 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
                     # so restart from the beginning.
                     return
 
-            if isinstance(obj, TimerHandle):
+            if isinstance(obj, asyncio.TimerHandle):
                 # A TimerHandle is added to the list of timers.
                 heapq.heappush(self._timers, obj)
                 return
@@ -732,29 +685,44 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
 
         # Don't go through the expensive nursery dance
         # if this is a sync function.
-        if getattr(obj, '_is_sync', True):
+        if isinstance(obj, AsyncHandle):
+            if hasattr(obj, '_context'):
+                obj._context.run(self._nursery.start_soon, obj._run, name=obj._callback)
+            else:
+                self._nursery.start_soon(obj._run, name=obj._callback)
+            await obj._started.wait()
+        else:
             if hasattr(obj, '_context'):
                 obj._context.run(obj._callback, *obj._args)
             else:
                 obj._callback(*obj._args)
-        else:
-            await self._nursery.start(obj._call_async)
 
     async def _main_loop_exit(self):
         """Finalize the loop. It may not be re-entered."""
         if self._closed:
             return
 
-        self.stop()
-        await self.wait_stopped()
+        with trio.CancelScope(shield=True):
+            # wait_stopped() will return once _main_loop() exits.
+            # stop() inserts a callback that will cause such, and
+            # _main_loop() doesn't block except to wait for new
+            # callbacks to be added, so this should be deadlock-proof.
+            self.stop()
+            await self.wait_stopped()
 
-        while True:
-            try:
-                await self._main_loop_one(no_wait=True)
-            except trio.WouldBlock:
-                break
-            except StopAsyncIteration:
-                pass
+            # Drain all remaining callbacks, even those after an initial
+            # call to stop(). This avoids deadlocks in some cases if
+            # work is submitted to the loop after the shutdown process
+            # starts. TODO: figure out precisely what this helps with,
+            # maybe find a better way. test_wrong_context_manager_order
+            # deadlocks if we remove it for now.
+            while True:
+                try:
+                    await self._main_loop_one(no_wait=True)
+                except trio.WouldBlock:
+                    break
+                except StopAsyncIteration:
+                    pass
 
         # Kill off unprocessed work
         self._cancel_fds()
@@ -818,43 +786,3 @@ class BaseTrioEventLoop(asyncio.SelectorEventLoop):
 
     def __exit__(self, *tb):
         raise RuntimeError("You need to use a sync loop, or 'async with open_loop()'.")
-
-    # Deprecated aliases #
-
-    run_future = staticmethod(
-        deprecated_alias(
-            "trio_asyncio.BaseTrioEventLoop.run_future",
-            _util.run_aio_future,
-            "0.10.0",
-            issue=38,
-        )
-    )
-
-    run_coroutine = deprecated_alias(
-        "trio_asyncio.BaseTrioEventLoop.run_coroutine",
-        run_aio_coroutine,
-        "0.10.0",
-        issue=38,
-    )
-
-    @deprecated("0.10.0", issue=38, instead="aio_as_trio(gen(*args))")
-    def wrap_generator(self, gen, *args):
-        return run_aio_generator(self, gen(*args))
-
-    @deprecated("0.10.0", issue=38, instead="aio_as_trio(aiter)")
-    def run_iterator(self, aiter):
-        return run_aio_generator(self, aiter)
-
-    @deprecated("0.10.0", issue=38, instead="aio_as_trio(proc)(*args)")
-    def run_asyncio(self, proc, *args):
-        from ._adapter import Asyncio_Trio_Wrapper
-        return Asyncio_Trio_Wrapper(proc, args=args, loop=self)
-
-    @deprecated("0.10.0", issue=38, instead="trio_as_aio(ctx)")
-    def wrap_trio_context(self, ctx):
-        from ._adapter import Trio_Asyncio_Wrapper
-        return Trio_Asyncio_Wrapper(ctx, loop=self)
-
-    @deprecated("0.10.0", issue=38, instead="trio_as_aio(proc)(*args)")
-    def run_trio(self, proc, *args):
-        return self.trio_as_future(proc, *args)
